/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.util.List;


import org.xsocket.DataConverter;





/**
 * http response
 *
 * @author grro@xlightweb.org
 */
public class HttpResponse extends AbstractHttpMessage implements IHttpResponse {
	
	private final IHttpResponseHeader responseHeader; 

    private final boolean isSimpleResponse;


	/**
	 * constructor 
	 * 
	 * @param responseHeader   the response header
	 */
	public HttpResponse(IHttpResponseHeader responseHeader) {
		this.responseHeader = responseHeader;
		isSimpleResponse = false;
	}	
	
	
	
	/**
	 * constructor 
	 * 
	 * @param responseHeader      the response header
	 * @param bodyDataSource      the response body
	 * 
	 * @throws IOException if an io exception occurs
	 */
	public HttpResponse(IHttpResponseHeader responseHeader, NonBlockingBodyDataSource bodyDataSource) throws IOException {
	    this(responseHeader, bodyDataSource, false);
	}	

	
	   /**
     * constructor 
     * 
     * @param responseHeader      the response header
     * @param bodyDataSource      the response body
     * 
     * @throws IOException if an io exception occurs
     */
	HttpResponse(IHttpResponseHeader responseHeader, NonBlockingBodyDataSource bodyDataSource, boolean isSimpleResponse) throws IOException {
        this.responseHeader = responseHeader;
        this.isSimpleResponse = isSimpleResponse;
        enhanceHeaderWithBodySize(bodyDataSource);
        
        setBodyDataSource(bodyDataSource);
    }   
	
	/**
	 * constructor 
	 * 
	 * @param responseHeader      the response header
	 * @param bodyDataSource      the response body
	 * 
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(String contentType, NonBlockingBodyDataSource bodyDataSource) throws IOException {
		responseHeader = new HttpResponseHeader(200, contentType);
		isSimpleResponse = false;
		enhanceHeaderWithBodySize(bodyDataSource);
		
		setBodyDataSource(bodyDataSource);
	}	

	
	/**
	 * constructor. The status will be set to 200 
	 * 
	 * @param responseHeader      the response header
	 * @param bodyDataSource      the response body
	 * 
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(String contentType, BlockingBodyDataSource bodyDataSource) throws IOException {
		responseHeader = new HttpResponseHeader(200, contentType);
		isSimpleResponse = false;
		enhanceHeaderWithBodySize(bodyDataSource.getUnderliyingBodyDataSource());
		
		setBodyDataSource(bodyDataSource.getUnderliyingBodyDataSource());
	}	
	

	/**
	 * constructor. The status will be set to 200
	 * 
	 * @param contentType    the content type 
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(String contentType, String body) throws IOException {
		this(200, contentType, body);
	}	

	
	
	/**
	 * constructor. The status will be set to 200
	 * 
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(String body) throws IOException {
		this(new HttpResponseHeader(200), body);
	}	


	/**
	 * constructor. The status will be set to 200
	 * 
	 * @param file           the file
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(File file) throws IOException {
		this(new HttpResponseHeader(200), new InMemoryBodyDataSource(AbstractHeader.DEFAULT_ENCODING, HttpUtils.newMultimodeExecutor(), HttpUtils.readFile(file)));

		setContentType(getContentTypeByFileExtension(file));
		setContentLength((int) file.length());
		removeHeader("Transfer-Encoding");
	}		
	

	 
	/**
	 * constructor 
	 * 
	 * @param status       the status 
	 */
	public HttpResponse(int status)  {
		responseHeader = new HttpResponseHeader(status);
		isSimpleResponse = false;
	}	

	
	
	/**
	 * constructor 
	 * 
	 * @param status       the status 
	 * @param contentType  the content type
	 * @param body         the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, String body) throws IOException {
		this(status, contentType, new ByteBuffer[] { DataConverter.toByteBuffer(body, HttpUtils.getMessageDefaultEncdoing()) }, HttpUtils.getMessageDefaultEncdoing());
	}	


	/**
	 * constructor 
	 * 
	 * @param status       the status 
	 * @param body         the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String body) throws IOException {
		this(new HttpResponseHeader(status), new ByteBuffer[] { DataConverter.toByteBuffer(body, AbstractHeader.DEFAULT_ENCODING)});
	}	

	/**
	 * constructor. The status will be set to 200
	 *  
	 * @param contentType    the content type
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(String contentType, byte[] body) throws IOException {
		this(200, contentType, body);
	}	



	/**
	 * constructor 
	 * 
	 * @param status         the status 
	 * @param contentType    the content type
	 * @param contentLength  the content length
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, int contentLength, NonBlockingBodyDataSource body) throws IOException {
		responseHeader = new HttpResponseHeader(status, contentType);
		responseHeader.setContentLength(contentLength);
		isSimpleResponse = false;
		
		setBodyDataSource(body);		
	}	
	
	
	
	/**
	 * constructor 
	 * 
	 * @param status         the status 
	 * @param contentType    the content type
	 * @param contentLength  the content length
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, int contentLength, BlockingBodyDataSource body) throws IOException {
		responseHeader = new HttpResponseHeader(status, contentType);
		responseHeader.setContentLength(contentLength);
		isSimpleResponse = false;
		
		setBodyDataSource(body.getUnderliyingBodyDataSource());		
	}	

	
	
	/**
	 * constructor 
	 * 
	 * @param status         the status 
	 * @param contentType    the content type
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, NonBlockingBodyDataSource body) throws IOException {
		if (body == null) {
			throw new NullPointerException("body has to be set");
		}
		responseHeader = new HttpResponseHeader(status, contentType);
		isSimpleResponse = false;
		enhanceHeaderWithBodySize(body);
		
		setBodyDataSource(body);		
	}	
	
	
	/**
	 * constructor 
	 * 
	 * @param status         the status 
	 * @param contentType    the content type
	 * @param body           the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, BlockingBodyDataSource body) throws IOException {
		responseHeader = new HttpResponseHeader(status, contentType);
		isSimpleResponse = false;
		enhanceHeaderWithBodySize(body.getUnderliyingBodyDataSource());
		
		setBodyDataSource(body.getUnderliyingBodyDataSource());		
	}	
	
	
	
	/**
	 * constructor 
	 * 
	 * @param status       the status 
	 * @param contentType  the content type
	 * @param body         the body
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, byte[] body) throws IOException {
		this(status, contentType, body, false);
	}	

	

    /**
     * constructor 
     * 
     * @param status       the status 
     * @param contentType  the content type
     * @param body         the body
     * @param compress     true, if the body should be compressed 
     * @throws IOException if an io exception occurs 
     */
    public HttpResponse(int status, String contentType, byte[] body, boolean compress) throws IOException {
        responseHeader = new HttpResponseHeader(status, contentType);
        isSimpleResponse = false;
        
        if (compress) {
            responseHeader.setHeader("Content-Encoding", "gzip");
        }
        
        int size = setBodyDataSource(body, responseHeader.getCharacterEncoding(), compress);
        responseHeader.setContentLength(size);
    }   

	

	/**
	 * constructor. The status will be set to 200 
	 * 
	 * @param contentType    the content type
	 * @param body           the body
	 * @throws IOException if an io exception occurs  
	 */
	public HttpResponse(String contentType, ByteBuffer[] body) throws IOException {
		this(200, contentType, body);
	}	
	
	
	/**
	 * constructor 
	 * 
	 * @param status       the status 
	 * @param contentType  the content type 
	 * @param body         the body 
	 * @throws IOException if an io exception occurs 
	 */
	public HttpResponse(int status, String contentType, ByteBuffer[] body) throws IOException {
		this(status, contentType, body, HttpUtils.getMessageDefaultEncdoing());			
	}	

	
	   
    /**
     * constructor 
     * 
     * 
     * @param status        the status 
     * @param contentType   the content type
     * @param body          the body 
     * @param encoding      the encoding
     * @throws IOException if an io exception occurs
     */
    HttpResponse(int status, String contentType, ByteBuffer[] body, String encoding) throws IOException {
       this(status, contentType, body, encoding, false);
    }   
    
    
	
	/**
	 * constructor 
	 * 
	 * 
	 * @param status        the status 
	 * @param contentType   the content type
	 * @param body          the body 
	 * @param encoding      the encoding
	 * @param compress      true, if the body should be compressed
	 * @throws IOException if an io exception occurs
	 */
	HttpResponse(int status, String contentType, ByteBuffer[] body, String encoding, boolean compress) throws IOException {
		responseHeader = new HttpResponseHeader(status, contentType + "; charset=" + encoding);
		isSimpleResponse = false;
		
		if (compress) {
		    responseHeader.setHeader("Content-Encoding", "gzip");
		}
		
		try {
			int size = setBodyDataSource(body, encoding, compress);
			responseHeader.setContentLength(size);
		} catch (UnsupportedEncodingException use) {
			throw new RuntimeException(use.toString());
		}
	}	
	
		
	/**
	 * constructor 
	 * 
	 * @param responseHeader   the response header 
	 * @param body             the body
	 * @throws IOException if an io exception occurs  
	 */
	public HttpResponse(IHttpResponseHeader responseHeader, String body) throws IOException {		
		this(responseHeader, body, false);
	}

	
	   
    /**
     * constructor 
     * 
     * @param responseHeader   the response header 
     * @param body             the body
     * @param compress         true, if the body should be compressed
     * @throws IOException if an io exception occurs  
     */
    public HttpResponse(IHttpResponseHeader responseHeader, String body, boolean compress) throws IOException {       
        this.responseHeader = responseHeader;
        isSimpleResponse = false;
        
        if (compress) {
            responseHeader.setHeader("Content-Encoding", "gzip");
        }
        
        try {
            int size = setBodyDataSource(body, responseHeader.getCharacterEncoding(), compress);
            responseHeader.setContentLength(size);
        } catch (UnsupportedEncodingException use) {
            throw new RuntimeException(use.toString());
        }
    }

	
	
	/**
     * constructor 
     * 
     * @param responseHeader   the response header
     * @param body             the body
     * @throws IOException if an io exception occurs  
     */
    public HttpResponse(IHttpResponseHeader responseHeader, byte[] body) throws IOException {
        this(responseHeader, body, false);     
    }
    
	
	/**
	 * constructor 
	 * 
	 * @param responseHeader   the response header
	 * @param body             the body
	 * @param compress         true, if the body should be compressed
	 * @throws IOException if an io exception occurs  
	 */
	public HttpResponse(IHttpResponseHeader responseHeader, byte[] body, boolean compress) throws IOException {
		this.responseHeader = responseHeader;
		isSimpleResponse = false;
		
		if (compress) {
		    responseHeader.setHeader("Content-Encoding", "gzip");
		}
		
		try {
			int size = setBodyDataSource(body, responseHeader.getCharacterEncoding(), compress);
			responseHeader.setContentLength(size);
		} catch (UnsupportedEncodingException use) {
			throw new RuntimeException(use.toString());
		}			
	}
	
	
	/**
     * constructor 
     * 
     * @param responseHeader   the response header
     * @param body             the body
     * @throws IOException if an io exception occurs  
     */
    public HttpResponse(IHttpResponseHeader responseHeader, List<ByteBuffer> body) throws IOException {
        this(responseHeader, body, false);
    }
    
    
    /**
     * constructor 
     * 
     * @param responseHeader   the response header
     * @param body             the body
     * @param compress         true, if the body should be compressed 
     * @throws IOException if an io exception occurs  
     */
    public HttpResponse(IHttpResponseHeader responseHeader, List<ByteBuffer> body, boolean compress) throws IOException {
        this(responseHeader, toArray(body), compress);
    }

    
	private static ByteBuffer[] toArray(List<ByteBuffer> buffers) {
	    if (buffers == null) {
	        return null;
	    } else {
	        return buffers.toArray(new ByteBuffer[buffers.size()]);
	    }
	}
    
	
	/**
	 * constructor 
	 * 
	 * @param responseHeader   the response header
	 * @param body             the body
	 * @throws IOException if an io exception occurs  
	 */
	public HttpResponse(IHttpResponseHeader responseHeader, ByteBuffer[] body) throws IOException {
		this(responseHeader, body, false);
	}

	
	/**
     * constructor 
     * 
     * @param responseHeader   the response header
     * @param body             the body
     * @param compress         true, if the body should be compressed 
     * @throws IOException if an io exception occurs  
     */
    public HttpResponse(IHttpResponseHeader responseHeader, ByteBuffer[] body, boolean compress) throws IOException {
        this.responseHeader = responseHeader;
        isSimpleResponse = false;
        
        if (body != null) {
            if (compress) {
                responseHeader.setHeader("Content-Encoding", "gzip");
            }
            
            try {
                int size = setBodyDataSource(body, responseHeader.getCharacterEncoding(), compress);
                responseHeader.setContentLength(size);
            } catch (UnsupportedEncodingException use) {
                throw new RuntimeException(use.toString());
            }
        }
    }

    
    
	
	
    
			
	/**
	 * {@inheritDoc}
	 */
	public final int getStatus() {
		return responseHeader.getStatus();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setStatus(int status) {
		responseHeader.setStatus(status);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getServer() {
		return responseHeader.getServer();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setServer(String server) {
		responseHeader.setServer(server);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setDate(String date) {
		responseHeader.setDate(date);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getDate() {
		return responseHeader.getDate();
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public final String getReason() {
		return responseHeader.getReason();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setReason(String reason) {
		responseHeader.setReason(reason);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getProtocol() {
		return responseHeader.getProtocol();
	}

	
	/**
	 * {@inheritDoc}
	 */
	public String getProtocolVersion() {
		return responseHeader.getProtocolVersion();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setProtocol(String protocol) {
		responseHeader.setProtocol(protocol);
	}
	

	/**
	 * returns the response header
	 *  
	 * @return the response header
	 */
	public final IHttpResponseHeader getResponseHeader() {
		return responseHeader;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public IHttpHeader getMessageHeader() {
		return getResponseHeader();
	}
	
	
	@Override
	public String toString() {
	    if (isSimpleResponse) {
	        return getBodyDataSource().toString();
	    } else {
	        return super.toString();
	    }
	}
}
